"""

Simple utility functions used in multiple modules.

"""

import sys
import subprocess
from colorama import init, Fore, Back, Style

USE_COLOR = True

args = None # Set globally from main() after command-line args are parsed.

# Available Colorama formatting constants are:
#
# Fore: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
# Back: BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE, RESET.
# Style: DIM, NORMAL, BRIGHT, RESET_ALL
#
# Style.RESET_ALL
# print(Fore.BLUE + 'some blue text')
# print(Fore.RED + 'some red text')
# print(Back.GREEN + 'and with a green background')
# print(Style.DIM + 'and in dim text')
# print(Style.RESET_ALL)
# print('back to normal now')

def init_color(parsed_args):
    """Initialize Colorama."""
    if not parsed_args.no_color:
        init(autoreset=False) # To avoid printing Style.RESET_ALL to clear.
    else:
        global USE_COLOR
        USE_COLOR = False

def print_color(color, *args, **kwargs):
    """Generic color printing."""
    if not USE_COLOR:
        print(*args, **kwargs)
    else:
        # For some reason this line below is needed or else BRIGHT text is not
        # turned off after printing, even though `autoreset=True` is set in init.
        #print(color, sep="", end="")
        #print(Style.RESET_ALL, sep="", end="") # This still fails to turn off sometimes...

        args = list(args)
        args[0] = Style.RESET_ALL + color + args[0]
        args[-1] = args[-1] + Style.RESET_ALL
        print(*args, **kwargs)

def print_info(*args, **kwargs):
    """Print out an ADB command with Colorama coloring."""
    print_color(Fore.GREEN, *args, **kwargs)

def input_query(arg):
    """Print out an ADB command with Colorama coloring."""
    if USE_COLOR:
        return input(Style.BRIGHT+Fore.CYAN + arg)
        #return input(Fore.GREEN + arg)
    else:
        return input(arg)

def print_cmd(*args, **kwargs):
    """Print out an ADB command with Colorama coloring."""
    print_color(Fore.BLUE, *args, **kwargs)

def print_error(*args, **kwargs):
    """Print out an ADB command with Colorama coloring."""
    print_color(Fore.RED, *args, **kwargs)

def print_warning(*args, **kwargs):
    """Print out an ADB command with Colorama coloring."""
    print_color(Fore.YELLOW, *args, **kwargs)

def query_yes_no(query_string, empty_default=None):
    """Query the user for a yes or no response.  The `empty_default` value can
    be set to a string to replace an empty response.  A "quit" response is
    taken to be the same as "no"."""
    yes_answers = {"Y", "y", "yes", "YES", "Yes"}
    no_answers = {"N", "n", "no", "NO", "No"}
    quit_answers = {"q", "Q", "quit", "QUIT", "Quit"}

    while True:
        response = input_query(query_string)
        response = response.strip()
        if empty_default is not None and response == "":
            return empty_default
        if not (response in yes_answers or response in no_answers or response in quit_answers):
            continue
        if response in yes_answers:
            return True
        return False # Must be a "no" or "quit" answer.

def run_local_cmd_blocking(cmd, *, print_cmd_str=False, print_cmd_prefix="", macro_dict={},
                           fail_on_nonzero_exit=True, capture_output=True):
    """Run a local system command.  If a string is passed in as `cmd` then
    `shell=True` is assumed.  If `macro_dict` is passed in then any dict key
    strings found as substrings of `cmd` will be replaced by their corresponding
    values.

    If `fail_on_nonzero_exit` is false then the return code is the first
    returned argument.  Otherwise only stdout and stderr are returned, assuming
    `capture_output` is true.

    Note that when `capture_output` is false the process output goes to the
    terminal as it runs, otherwise it doesn't."""
    shell = False
    if isinstance(cmd, str):
        shell = True # Run as shell cmd if a string is passed in.
        for key, value in macro_dict.items():
            cmd = cmd.replace(key, value)
        cmd_string = cmd
    else:
        for key, value in macro_dict.items():
            cmd = [s.replace(key, value) for s in cmd]
        cmd_string = " ".join(cmd)

    if print_cmd_str:
        cmd_string = "\n" + print_cmd_prefix + cmd_string
        print_cmd(cmd_string)

    completed_process = subprocess.run(cmd, capture_output=capture_output, shell=shell,
                                       check=False, encoding="utf-8")

    if fail_on_nonzero_exit and completed_process.returncode != 0:
        print_error("\nError, nonzero exit running system command, exiting...", file=sys.stderr)
        sys.exit(1)

    if capture_output:
        if fail_on_nonzero_exit:
            return completed_process.stdout, completed_process.stderr
        return completed_process.returncode, completed_process.stdout, completed_process.stderr
    if not fail_on_nonzero_exit:
        return completed_process.returncode

def indent_lines(string, n=4):
    """Indent all the lines in a string by n spaces."""
    string_list = string.splitlines()
    string_list = [" "*n + i for i in string_list]
    return "\n".join(string_list)




#!/usr/bin/env python3
"""

Usage: recdroidvid.py

Be sure to install scrcpy and set phone to allow for ADB communication over
USB.  See the video recording notes in ardour directory for details.

Currently requires these programs to be installed:
    scrcpy
    ffprobe, to print information about videos
    ffmpeg, for audio extraction when that option is selected
    mpv, for previewing when that option is selected

    sudo apt install scrcpy ffmpeg mpv

"""

import sys
import os
from time import sleep
import subprocess
import datetime
import threading

from .utility_functions import print_info, print_error, print_warning
from .utility_functions import init_color, query_yes_no, indent_lines, run_local_cmd_blocking
from . import utility_functions

from .settings_and_options import (parse_command_line, DETECT_JACK_CMD,
                USE_SCREENRECORD, RECORD_DETECTION_METHOD, SYNC_DAW_SLEEP_TIME,
                VIDEO_FILE_EXTENSION, QUERY_EXTRACT_AUDIO, QUERY_PREVIEW_VIDEO,
                EXTRACTED_AUDIO_EXTENSION, POSTPROCESS_VIDEOS,
                POSTPROCESSING_CMD)

from . import adb_commands as adb
from .adb_commands import ADBException

args = None # Set globally from main() after command-line args are parsed.

VERSION = "0.1.1"

DEBUG = False

#
# Local machine startup functions.
#

def print_startup_message():
    """Print out the initial greeting message."""
    print_info(f"{'='*78}")
    print_info(f"\nrecdroidvid, version {VERSION}")
    print_info(f"\n{'='*78}")

def print_debug(*args, **kwargs):
    """Print debugging message only if DEBUG is set true in module scope."""
    # TODO: Maybe add a prefix option to print DEBUG first.
    if DEBUG:
        print(*args, **kwargs)

def detect_if_jack_running():
    """Determine if the Jack audio system is currently running; return true if it is."""
    errorcode, stdout, stderr = run_local_cmd_blocking(DETECT_JACK_CMD, fail_on_nonzero_exit=False)
    if errorcode == 0:
        return True
    return False
    ## This older code works, looking at ps output, but is probably less robust.
    #ps_output, stderr = run_local_cmd_blocking(["ps", "-ef"])
    #ps_output = ps_output.splitlines()
    #for p in ps_output:
    #    #print(p)
    #    for pname in DETECT_JACK_PROCESS_NAMES:
    #        if pname in p:
    #            return True
    #return False

#
# Functions for syncing with DAW.
#

def raise_daw_in_window_stack():
    """Run the command to raise the DAW in the window stack."""
    print_info("\nRaising DAW to top of Window stack.")
    # Allow the command to fail, but issue a warning.
    returncode, stdout, stderr = run_local_cmd_blocking(args.raise_daw_to_top_cmd[0],
                                     print_cmd_str=True, print_cmd_prefix="SYSTEM: ",
                                     fail_on_nonzero_exit=False)
    if returncode != 0:
        print_warning("\nWARNING: Nonzero exit status running the raise-DAW command.", file=sys.stderr)
    return returncode

def is_daw_running():
    """Return true or false as to whether the DAW is running."""
    returncode, stdout, stderr = run_local_cmd_blocking(args.is_daw_running_cmd[0],
                                                               fail_on_nonzero_exit=False)
    if returncode != 0:
        print_debug(f"\nDEBUG: DAW not detected as running.")
        return False
    print_debug(f"\nDEBUG: DAW detected as running.")
    return True

def start_daw_recording():
    """Start the recording transport state of the DAW.  Used to sync with recording."""
    if not is_daw_running():
        print_warning("WARNING: DAW is not detected as running, not toggling transport.",
                file=sys.stderr)
        return
    returncode, stdout, stderr = run_local_cmd_blocking(args.start_daw_recording_cmd[0],
                                           print_cmd_str=True, print_cmd_prefix="SYSTEM: ",
                                                               fail_on_nonzero_exit=False)
    if returncode !=0:
        print_warning("WARNING: Nonzero exit status running the start-daw-recording command.", file=sys.stderr)
    if args.raise_daw_on_transport_toggle:
        raise_returncode = raise_daw_in_window_stack()

def stop_daw_transport():
    """Stop the DAW transport.  Used to sync with recording."""
    if not is_daw_running():
        print_warning("WARNING: DAW is not detected as running, not toggling transport.",
                file=sys.stderr)
        return
    returncode, stdout, stderr = run_local_cmd_blocking(args.stop_daw_transport_cmd[0],
                                           print_cmd_str=True, print_cmd_prefix="SYSTEM: ",
                                                               fail_on_nonzero_exit=False)
    if returncode !=0:
        print_warning("WARNING: Nonzero exit status running the stop-daw-transport command.", file=sys.stderr)
    if args.raise_daw_on_transport_toggle:
        raise_returncode = raise_daw_in_window_stack()

def add_mark_in_daw():
    """Create a new mark in the DAW when recording is started."""
    if not is_daw_running():
        print_warning("WARNING: DAW is not detected as running, not adding a mark.", file=sys.stderr)
        return
    print_info(f"\nAdding a new mark in the DAW.")
    run_local_cmd_blocking(args.add_daw_mark_cmd[0], print_cmd_str=True,
                           print_cmd_prefix="SYSTEM: ")

sync_daw_stop_flag = False # Flag to signal the DAW sync thread to stop.

def video_is_recording_on_device():
    """Function to detect when video is recording on the Android device, returns
    true or false."""
    if RECORD_DETECTION_METHOD == "directory size increasing":
        return adb.directory_size_increasing(args.camera_save_dir[0],
                                             wait_secs=1)
    if RECORD_DETECTION_METHOD == ".pending filename prefix":
        return adb.pending_video_file_exists(args.camera_save_dir[0])

    print_error(f"Error in recdroidvid setting: Unrecognized RECORD_DETECTION_METHOD:"
          f"\n   '{RECORD_DETECTION_METHOD}'", file=sys.stderr)
    sys.exit(1)

def sync_daw_transport_bg_process(stop_flag_fun):
    """Start the DAW transport when video recording is detected on the Android
    device.  Meant to be run as a thread or via multiprocessing to execute at the
    same time as the scrcpy monitor."""
    daw_transport_rolling = False
    while True:
        vid_recording = video_is_recording_on_device()
        if not daw_transport_rolling and vid_recording: # Start DAW recording transport.
            print_debug(f"\n{daw_transport_rolling=}   {vid_recording=}")
            print_info("\nStarting (toggling) DAW transport.")
            if args.add_daw_mark_on_transport_start:
                add_mark_in_daw()
            start_daw_recording()
            daw_transport_rolling = True
        if daw_transport_rolling and not vid_recording: # Stop DAW recording transport.
            print_debug(f"\n{daw_transport_rolling=}   {vid_recording=}")
            print_info("\nStopping (toggling) DAW transport.")
            stop_daw_transport()
            daw_transport_rolling = False
        if stop_flag_fun():
            break
        sleep(SYNC_DAW_SLEEP_TIME)

def sync_daw_transport_with_video_recording():
    """Start up the background process to sync the DAW transport when recording
    starts or stops are detected on the mobile device."""
    # To use threading instead, set a stop flag as in one of the answers here:
    # https://stackoverflow.com/questions/323972/is-there-any-way-to-kill-a-thread
    proc = threading.Thread(target=sync_daw_transport_bg_process,
                                   args=(lambda: sync_daw_stop_flag,))
    proc.daemon = True # This is so the thread always dies when the main program exits.
    proc.start()
    return proc

def sync_daw_process_kill(proc):
    """Kill the DAW syncing process and reclaim resources."""
    global sync_daw_stop_flag
    sync_daw_stop_flag = True
    proc.join()
    sync_daw_stop_flag = False # Reset for next time.

#
# Recording and monitoring functions.
#

def start_screenrecording():
    """Start screenrecording via the ADB `screenrecord` command.  This process is run
    in the background.  The PID is returned along with the video pathname."""
    # CODE DEPRECATED AND NOW UNTESTED!!!!
    video_out_basename = args.video_file_prefix[0]
    video_out_pathname =  os.path.join(args.camera_save_dir[0], f"{video_out_basename}.mp4")
    tmp_pid_path = f"zzzz_screenrecord_pid_tmp"
    adb.ls(os.path.dirname(video_out_pathname)) # DOESNT DO ANYTHING?? DEBUG??

    adb.adb(f"adb shell screenrecord {video_out_pathname} & echo $! > {tmp_pid_path}")

    sleep(1)
    with open(tmp_pid_path, "r", encoding="utf-8") as f:
        pid = f.read()
    os.remove(tmp_pid_path)
    sleep(10)

    # NOTE, below takes --size, but messes it up, density??
    #adb shell screenrecord --size 720x1280 /storage/emulated/0/DCIM/OpenCamera/$1.mp4 &
    return pid, video_out_pathname

def start_screen_monitor():
    """Run the scrcpy program as a screen monitor, blocking until it is shut down."""
    # Note cropping is width:height:x:y  [currently FAILS as below, video comes out
    # broken too]
    #
    # https://old.reddit.com/r/sidequest/comments/ed9xzc/what_crop_number_should_i_enter_in_scrcpy_if_i/
    #    The syntax is: --crop width:height:x:y. So if you pass 1920:1080:1440:720, you
    #    want a video 1920×1080, starting at (1440,720) and ending at (3360, 1800)
    #    [Last coords are the offsets added to the first coords, as an ordered pair.]
    #
    # Actual is 1600x720 (in landscape) for phone screen, so 720:1280:0:160  (260 before, worked mostly...)
    #
    # Note that capturing full 1600x720 is possible, but below cropped to 16:9.

    # NOTE that lock-video-orientation is gone in recent scrcpy, but you can
    # pass an @ symbol in different ways to --orientation (or
    # --display-orientation and --capture-orientation).
    #

    # Uncropped.
    #scrcpy --record=$1.mp4 --record-format=mp4 --orientation=0 --stay-awake --disable-screensaver --video-buffer=50 --crop 720:1280:0:160 # --crop 720:1600:0:0

    # Cropped to 16:9.
    #scrcpy --record=$1.mp4 --record-format=mp4 --orientation=0 --stay-awake --disable-screensaver --video-buffer=50 --crop 720:1280:0:320 # --crop 720:1600:0:0

    print_info("\nStarting the scrcpy program.")

    scrcpy_cmd = args.scrcpy_cmd[0]
    #window_title_str = f"video file prefix: {args.video_file_prefix}" # TODO, this caused cmd line error, below fixes.
    window_title_str = f"{args.video_file_prefix}"
    run_local_cmd_blocking(scrcpy_cmd, print_cmd_str=True, print_cmd_prefix="SYSTEM: ",
                           macro_dict={"RDV_SCRCPY_TITLE": window_title_str},
                           capture_output=False)

def start_monitoring_and_button_push_recording():
    """Emulate a button push to start and stop recording."""
    # Get a snapshot of save directory before recording starts.
    before_ls = adb.ls(args.camera_save_dir[0], extension_whitelist=[VIDEO_FILE_EXTENSION])

    if args.autorecord:
        adb.tap_camera_button()

    if args.sync_daw_transport_with_video_recording:
        proc = sync_daw_transport_with_video_recording()

    start_screen_monitor() # This blocks until the screen monitor is closed.

    # If the user just shut down scrcpy while recording video, stop the recording.
    if adb.directory_size_increasing(args.camera_save_dir[0]):
        adb.tap_camera_button() # Presumably still recording; turn off the camera.
        #if args.sync_daw_transport_with_video_recording: # Now BG thread is still running to stop DAW transport.
        #    stop_daw_transport() # Presumably the DAW transport is still rolling.
        while adb.directory_size_increasing(args.camera_save_dir[0]):
            print_info("Waiting for save directory to stop increasing in size...")
            sleep(1)

    if args.sync_daw_transport_with_video_recording:
        sync_daw_process_kill(proc)

    # Get a final snapshot of save directory after recording is finished.
    after_ls = adb.ls(args.camera_save_dir[0], extension_whitelist=[VIDEO_FILE_EXTENSION])

    new_video_files = [f for f in after_ls if f not in before_ls]
    new_video_paths = [os.path.join(args.camera_save_dir[0], v) for v in new_video_files]
    return new_video_paths

def generate_video_name(video_number, pulled_vid_name):
    """Generate the name to rename a pulled video to."""
    if args.date_and_time_in_video_name:
        date_time_string = datetime.datetime.now().strftime('%Y-%m-%d_%H.%M.%S_')
    else:
        date_time_string = ""
    new_vid_name = f"{args.video_file_prefix}_{video_number:02d}_{date_time_string}{pulled_vid_name}"
    return new_vid_name

def monitor_record_and_pull_videos(video_start_number):
    """Record a video on the Android device and pull the resulting file."""
    if USE_SCREENRECORD: # NOTE: This method is no longer tested, may be removed.
        recorder_pid, video_path = start_screenrecording()
        start_screen_monitor()
        run_local_cmd_blocking(f"kill {recorder_pid}")
        video_path = pull_and_delete_file(video_path)
        return [video_path]

    # Use the method requiring a button push on phone, emulated or actual.
    video_paths = start_monitoring_and_button_push_recording()
    new_video_paths = []
    sleep(5) # Make sure video files have time to finish writing and close.
    for count, vid in enumerate(video_paths):
        pulled_vid = pull_and_delete_file(vid) # Note file always written to CWD for now.
        sleep(0.3)
        new_vid_name = generate_video_name(count+video_start_number, pulled_vid)
        print_info(f"\nSaving (renaming) video file as\n   {new_vid_name}")
        os.rename(pulled_vid, new_vid_name)
        new_video_paths.append(new_vid_name)
    return new_video_paths

#
# Video postprocessing functions.
#

def pull_and_delete_file(pathname):
    """Pull the file at the pathname and delete the remote file.  Returns the
    path of the extracted video."""
    print_info("\nPulling recorded video(s) from the phone, then deleting them there.")

    # Pull.
    adb.adb(f"adb pull {pathname}")

    # Delete.
    sleep(4)
    adb.adb(f"adb shell rm {pathname}")
    sleep(1)
    adb.adb(f"adb -d shell am broadcast -a android.intent.action.MEDIA_SCANNER_SCAN_FILE -d file:{pathname}")
    return os.path.basename(pathname)

PREVIEW_WINDOW_ALWAYS_ON_TOP = False # TODO, possible feature.  But preview blocking messes it up...
SET_ACTIVE_WINDOW_ALWAYS_ON_TOP_CMD = ["wmctrl", "-r", ":ACTIVE:", "-b", "toggle,above"]

def preview_video(video_path):
    """Run a preview of the video at `video_path`."""
    if not (args.preview_video or QUERY_PREVIEW_VIDEO):
        return
    if QUERY_PREVIEW_VIDEO and not query_yes_no("\nRun preview? "):
        return

    print_info("\nRunning preview...")
    if detect_if_jack_running():
        print_info("\nDetected jack audio running.")
        preview_cmd = args.preview_video_cmd_jack[0] + f" {video_path}"
    else:
        print_info("\nDid not detect jack audio running.")
        preview_cmd = args.preview_video_cmd[0] + f" {video_path}"

    run_local_cmd_blocking(preview_cmd, print_cmd_str=True, capture_output=False,
                       print_cmd_prefix="SYSTEM: ",
                       macro_dict={"RDV_PREVIEW_FILENAME": os.path.basename(video_path)})

    if PREVIEW_WINDOW_ALWAYS_ON_TOP:
        run_local_cmd_blocking(SET_ACTIVE_WINDOW_ALWAYS_ON_TOP_CMD, print_cmd_str=True,
                               capture_output=False)

def extract_audio_from_video(video_path):
    """Extract the audio from a video file, of the type with the given extension."""
    if not ((args.audio_extract or QUERY_EXTRACT_AUDIO) and os.path.isfile(video_path)
                                                   and not USE_SCREENRECORD):
        return
    if QUERY_EXTRACT_AUDIO and not query_yes_no("\nExtract audio from video? "):
        return

    dirname, basename = os.path.split(video_path)
    root_name, video_extension = os.path.splitext(basename)
    output_audio_path = os.path.join(dirname, root_name + EXTRACTED_AUDIO_EXTENSION)
    print_info(f"\nExtracting audio to file: '{output_audio_path}'")
    # https://superuser.com/questions/609740/extracting-wav-from-mp4-while-preserving-the-highest-possible-quality
    cmd = f"ffmpeg -i {video_path} -map 0:a {output_audio_path} -loglevel quiet"
    run_local_cmd_blocking(cmd, print_cmd_str=True, print_cmd_prefix="SYSTEM: ",
                           capture_output=False)
    print_info("\nAudio extracted.")

def postprocess_video_file(video_path):
    """Run a postprocessing algorithm on the video file at `video_path`."""
    if not POSTPROCESS_VIDEOS or not os.path.isfile(video_path):
        return
    postprocess_cmd = POSTPROCESSING_CMD + [f"{video_path}"]
    run_local_cmd_blocking(postprocess_cmd, print_cmd_str=True, print_cmd_prefix="SYSTEM: ",
                           capture_output=False)

def print_info_about_pulled_video(video_path):
    """Print out some information about the resolution, etc., of a video."""
    # To get JSON: ffprobe -v quiet -print_format json -show_format -show_streams "lolwut.mp4" > "lolwut.mp4.json"
    # In Python search for examples or use library: https://docs.python.org/3/library/json.html
    cmd = (f"ffprobe -pretty -show_format -v error -show_entries"
           f" stream=codec_name,width,height,duration,size,bit_rate"
           f" -of default=noprint_wrappers=1 {video_path} | grep -v 'TAG:'")
    print_info("\nRunning ffprobe on saved video file:")
    stdout, stderr = run_local_cmd_blocking(cmd)
    print(indent_lines(stdout, 4))
    if stderr:
        print(indent_lines(stderr, 4))

#
# High-level functions.
#

def startup_device_and_run(video_start_number):
    """Main script functionality."""
    # Note first adb call must raise exception, not just call sys.exit.
    adb.device_sleep(exit_on_error=False) # Get a consistent starting state for repeatability.
    adb.device_wakeup()
    adb.unlock_screen()
    adb.open_video_camera()
    if args.raise_daw_on_camera_app_open:
        raise_daw_in_window_stack()

    video_paths = monitor_record_and_pull_videos(video_start_number)
    adb.device_sleep() # Put the device to sleep after use.

    for vid in video_paths:
        print_info(f"\n{'='*12} {vid} {'='*30}")
        print_info_about_pulled_video(vid)
        preview_video(vid)
        extract_audio_from_video(vid)
        postprocess_video_file(vid)

    video_end_number = video_start_number + len(video_paths) - 1
    return video_end_number

def main():
    """Outer loop over invocations of the scrcpy screen monitor."""
    # Parse the command-line arguments and set the args global for modules that use it.
    global args
    args = parse_command_line()
    utility_functions.args = args
    adb.args = args

    # General init stuff.
    init_color(args)
    video_start_number = args.numbering_start[0]
    print_startup_message()

    count = 0
    device_found = True
    while True:
        if device_found == False or count == 0 and args.wait_loop:
            cont = query_yes_no(f"\nRecdroidvid wait loop, continue?"
                                f" [ynq enter=y]: ", empty_default="y")
            device_found = True
            if not cont:
                print_info("\nExiting recdroidvid.")
                return
        count += 1

        # Loop until device is detected.
        try:
            video_end_number = startup_device_and_run(video_start_number)
        except ADBException:
            count -= 1
            device_found = False
            continue

        video_start_number = video_end_number + 1
        if not args.loop:
            break
        cont = query_yes_no(f"\nFinished recdroidvid loop {count}, continue?"
                            f" [ynq enter=y]: ", empty_default="y")
        if not cont:
            break
    print_info("\nExiting recdroidvid.")

if __name__ == "__main__":

    main()




"""

Fixed options and command-line options.

Note: The keycodes for xdotool.
   https://gitlab.com/cunidev/gestures/-/wikis/xdotool-list-of-key-codes

"""

# NOTE: To find this path, look at the info from an OpenCamera video saved on the phone.
OPENCAMERA_SAVE_DIR = "/storage/emulated/0/DCIM/OpenCamera/" # Where OpenCamera writes video.
OPENCAMERA_PACKAGE_NAME = "net.sourceforge.opencamera" # Look in URL of PlayStore page to find.

VIDEO_FILE_EXTENSION = ".mp4"

# The default command-line args passed to scrcpy.
# Note the title macro is substituted-in later.
SCRCPY_CMD_DEFAULT = ["scrcpy", "--stay-awake",
                                "--disable-screensaver",
                                "--video-buffer=20",
                                "--window-title=RDV_SCRCPY_TITLE",
                                "--always-on-top",
                                "--orientation=0",
                                "--max-size=1200",]

BASE_VIDEO_PLAYER_CMD = ["mpv", "--loop=inf",
                                "--autofit=1080", # Set the width of displayed video.
                                #"--geometry=50%:70%", # Set initial position on screen.
                                #"--autosync=30",
                                "--cache=yes",
                                "--osd-duration=200",
                                "--osd-bar-h=0.5",
                                #"--really-quiet", # Turn off when debugged; masks errors.
                                f"--title='{'='*8} VIDEO PREVIEW: RDV_PREVIEW_FILENAME {'='*40}'"]

VIDEO_PLAYER_CMD = BASE_VIDEO_PLAYER_CMD # + ["--ao=sdl"]
VIDEO_PLAYER_CMD_JACK = VIDEO_PLAYER_CMD + ["--ao=jack"]

#DETECT_JACK_PROCESS_NAMES = [" qjackctl "] # Search `ps -ef` for these to detect Jack running.
DETECT_JACK_CMD = "jack_lsp -A" # Cmd to detect jack; nonzero return code if not running.

QUERY_PREVIEW_VIDEO = False # Ask before previewing video.
QUERY_EXTRACT_AUDIO = False # Ask before extracting AUDIO file.

EXTRACTED_AUDIO_EXTENSION = ".wav"

IS_DAW_RUNNING_CMD = 'xdotool search --onlyvisible --class Ardour'
START_DAW_RECORDING_CMD = 'xdotool key --window "$(xdotool search --onlyvisible --class Ardour | head -1)" shift+space'
STOP_DAW_TRANSPORT_CMD = 'xdotool key --window "$(xdotool search --onlyvisible --class Ardour | head -1)" space'
#TOGGLE_DAW_TRANSPORT_CMD = 'xdotool windowactivate "$(xdotool search --onlyvisible --class Ardour | head -1)"'

ADD_DAW_MARK_CMD = 'xdotool key --window "$(xdotool search --onlyvisible --class Ardour | head -1)" Tab'

RAISE_DAW_TO_TOP_CMD = "xdotool search --onlyvisible --class Ardour windowactivate %@"

SYNC_DAW_SLEEP_TIME = 4 # Lag between video on/off & DAW transport sync (load/time tradeoff)

RECORD_DETECTION_METHOD = "directory size increasing" # More general but requires two calls.
# Below started failing, starting on open before video actually started... was a .pending file
# there from before, might be related...
#RECORD_DETECTION_METHOD = ".pending filename prefix" # May be specific to OpenCamera implemetation.

# This option records with the ADB screenrecord command.  It is limited to the
# screen's resolution(?) and 3 minutes, with no sound.  It is no longer tested
# and will be removed at some point.  https://stackoverflow.com/questions/21938948/
USE_SCREENRECORD = False # DEPRECATED.

POSTPROCESS_VIDEOS = False
POSTPROCESSING_CMD = [] # Enter cmd as separate string arguments.

RECDROIDVID_PYTHON_RC_FILENAME = ".recdroidvid_rc.py"

import sys
import os
import argparse
#import ast
import shutil
import tempfile
from .utility_functions import print_info, print_error, print_warning

def fmt(text):
    """Format text for nicer-looking help messages."""
    # https://docs.python.org/3/library/textwrap.html
    import textwrap
    text_lines = textwrap.wrap(textwrap.dedent(text), width=70)
    print(text_lines)
    text = "\n".join(text_lines) + "\n\n"
    print(text)
    return text

def parse_command_line():
    """Create and return the argparse object to read the command line."""

    parser = argparse.ArgumentParser(
                        prog="recdroidvid",
                        #formatter_class=argparse.RawTextHelpFormatter,
                        #formatter_class=argparse.RawDescriptionHelpFormatter,
                        description=

                        """Record a video on mobile via ADB and pull result.  All config
                        options can be set in a file `.recdroidvid_rc.py`.  The file is
                        evaluated and the list `rdv_options` in the file is used as the
                        options list.  See the example config file.""")

    parser.add_argument("video_file_prefix", type=str, nargs="?", metavar="PREFIXSTRING",
                        default="rdv", help="""The basename or prefix of the pulled video
                        file.  Whether name or prefix depends on the method used to
                        record.""")

    parser.add_argument("--scrcpy-cmd", "-y", type=str, nargs=1, metavar="CMD-STRING",
                        default=[" ".join(SCRCPY_CMD_DEFAULT)], help="""The command,
                        including arguments, to be used to launch the scrcpy program.
                        Otherwise a default version is used with some common arguments.
                        Note that the string `--window-title=RDV_SCRCPY_TITLE` can be used
                        to substitute-in a more descriptive title for the window.""")

    parser.add_argument("--numbering-start", "-n", type=int, nargs=1, metavar="INTEGER",
                        default=[1], help="""The number at which to start numbering
                        pulled videos.  The number is currently appended to the user-defined
                        prefix and defaults to 1.  Allows for restarting and continuing
                        a naming sequence across invocations of the program.""")

    parser.add_argument("--loop", "-l", action="store_true", help="""
                        Loop the recording, querying between invocations of `scrcpy` as to
                        whether or not to continue.  This allows for shutting down the
                        scrcpy display to save both local CPU and remote device memory
                        (videos are downloaded and deleted from the device at the end of
                        each loop), but then restarting with the same options.  Video
                        numbering (as included in the filename) is automatically incremented
                        over all the videos, across loops.""")

    parser.add_argument("--wait-loop", "-w", action="store_true", help="""
                        The '--loop' option always starts the scrcpy video monitor
                        immediately on the first loop.  This option delays the action until
                        the user responds to a query.  This avoids the CPU cost of scrcpy if
                        you are not planning to start video recording right away.  This
                        option implies the '--loop' option.""")

    parser.add_argument("--autorecord", "-a", action="store_true",
                        default=False, help="""Automatically start recording when the scrcpy
                        monitor starts up.""")

    parser.add_argument("--preview-video", "-p", action="store_true",
                        default=False, help="""Preview each video that is downloaded.
                        Currently uses the mpv program.""")

    parser.add_argument("--preview-video-cmd", type=str, nargs=1, metavar="CMD-STRING",
                        default=[" ".join(VIDEO_PLAYER_CMD)], help="""The command used to
                        invoke a movie player to view the preview.  The default
                        uses the mpv movie viewer.  The string 'RDV_PREVIEW_FILENAME', if
                        present in the command, will be replaced with the title of the
                        video being previewed.""")

    parser.add_argument("--preview-video-cmd-jack", type=str, nargs=1, metavar="CMD-STRING",
                        default=[" ".join(VIDEO_PLAYER_CMD_JACK)], help="""The command used to
                        invoke a movie player to view the preview when the jack audio
                        system is detected to be running.  The default uses the mpv
                        movie viewer.  The string 'RDV_PREVIEW_FILENAME', if
                        present in the command, will be replaced with the title of the
                        video being previewed.""")

    parser.add_argument("--date-and-time-in-video-name", "-t", action="store_true",
                        default=False, help="""Include the date and time in the video names
                        in a readable format.""")

    parser.add_argument("--sync-daw-transport-with-video-recording", "-s", action="store_true",
                        default=False, help="""Start the DAW transport when
                        video recording is detected on the mobile device.  May increase
                        CPU loads on the computer and the mobile device.""")

    parser.add_argument("--start-daw-recording-cmd", type=str, nargs=1, metavar="CMD-STRING",
                        default=[START_DAW_RECORDING_CMD], help="""A system command to start
                        DAW recording.  Used when the `--sync-to-daw` option is chosen.  The
                        default uses xdotool to send a shift-space character to Ardour.""")

    parser.add_argument("--stop-daw-transport-cmd", type=str, nargs=1, metavar="CMD-STRING",
                        default=[STOP_DAW_TRANSPORT_CMD], help="""A system command to stop the
                        DAW transport.  Used when the `--sync-to-daw` option is chosen.  The
                        default uses xdotool to send a space character to Ardour.""")

    parser.add_argument("--add-daw-mark-on-transport-start", "-m", action="store_true",
                        help="""Whether to add a mark in the DAW when the
                        transport starts, to help in syncing with the video.""")

    parser.add_argument("--add-daw-mark-cmd", type=str, nargs=1, metavar="CMD-STRING",
                        default=[ADD_DAW_MARK_CMD], help="""A system command to add
                        a mark to the DAW at the playhead.  The default uses xdotool to send
                        a tab character to Ardour.""")

    parser.add_argument("--raise-daw-on-camera-app-open", "-q", action="store_true",
                        help="""Raise the DAW to the top
                        of the window stack when the camara app is opened on the mobile device.
                        Works well when scrcpy is also passed the `--always-on-top` option.""")

    parser.add_argument("--raise-daw-on-transport-toggle", "-r", action="store_true",
                        default=False, help= """Raise the DAW to the top of the window
                        stack whenever the DAW transport is toggled by the `--sync-to-daw`
                        option.  Works well when scrcpy is also passed the
                        `--always-on-top` option.""")

    parser.add_argument("--raise-daw-to-top-cmd", type=str, nargs=1, metavar="CMD-STRING",
                        default=[RAISE_DAW_TO_TOP_CMD], help="""A system command to raise the
                        DAW windows to the top of the window stack.  Used when either of the
                        `--raise_daw_on_camera_app_open` or `--raise-daw-on-transport-toggle`
                        options are selected.  The default uses xdotool to activate any Ardour
                        windows.""")

    parser.add_argument("--is-daw-running-cmd", type=str, nargs=1, metavar="CMD-STRING",
                        default=[IS_DAW_RUNNING_CMD], help="""A system command to test if
                        the DAW is actually running.  A zero return code means it is, and
                        a nonzero return code means it isn't.""")

    parser.add_argument("--audio-extract", "-e", action="store_true", default=False,
                        help="""Extract a separate audio file (currently always a WAV file)
                        from each video.""")

    parser.add_argument("--camera-save-dir", "-d", type=str, nargs=1, metavar="DIRPATH",
                        default=[OPENCAMERA_SAVE_DIR], help="""The directory on the remote
                        device where the camera app saves videos.  Record a video and look
                        at the information about the video to find the path.   Defaults
                        to the OpenCamera default save directory.""")

    parser.add_argument("--camera-package-name", "-c", type=str, nargs=1, metavar="PACKAGENAME",
                        default=[OPENCAMERA_PACKAGE_NAME], help="""The Android package name of
                        the camera app.  Defaults to "net.sourceforge.opencamera", the
                        OpenCamera package name.  Look in the URL of the app's PlayStore
                        web site to find this string.""")

    parser.add_argument("--config-conditional", type=str, nargs=1, metavar="STRING",
                        default=["default"], help="""The `.recdroidvid_rc.py` config file
                        contains interpreted Python code, so conditionals can be set for different
                        use-cases.  This option allows one to set a string value from the command
                        line which can then be used to choose a case in the config file.  To set
                        such a variable, pass the value to this option.  The default value is
                        the string "default".  To access this variable, use
                        `from recdroidvid import config_conditional` at the top of the config
                        file.
                        """)

    parser.add_argument("--no-color", action="store_true", default=False, help="""
                        Do not use color highlighting on the terminal output.""")

    # Set the variable for the `--config-conditional` option, based ONLY on cmdline args.
    parsed_cmdline_only_args = parser.parse_args()
    import recdroidvid as rdv
    if parsed_cmdline_only_args.config_conditional:
        rdv.config_conditional = parsed_cmdline_only_args.config_conditional[0]
    else:
        rdv.config_conditional = "default"

    # Now parse the commandline again, with the combined config args and cmdline args.
    rc_file_args = read_python_rc_file()
    #rc_file_args = read_rc_file()
    combined_args = rc_file_args + sys.argv[1:]
    parsed_args = parser.parse_args(args=combined_args)

    if parsed_args.wait_loop: # The wait-loop option implies loop.
        parsed_args.loop = True

    global args
    args = parsed_args
    return parsed_args

#def read_rc_file():
#    """Read and parse the ~/.recdroidvid_rc file."""
#    rc_path = os.path.abspath(os.path.expanduser("~/.recdroidvid_rc"))
#    if not os.path.isfile(rc_path):
#        return []
#
#    with open(rc_path, "r", encoding="utf-8") as f:
#        text = f.read()
#
#    try:
#        args_list = ast.literal_eval("[" + text + "]")
#    except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError) as e:
#        print(f"\nError parsing ~/.recdroidvid_rc:\n   ", e, file=sys.stderr)
#        print("\nExiting.")
#        sys.exit()
#
#    for a in args_list: # Make sure everything evaluated as a string.
#        if not isinstance(a, str):
#            print(f"\nError parsing ~/.recdroidvid_rc: The option or value '{a}' is not "
#                    "a quoted string.", file=sys.stderr)
#            print("\nExiting.")
#            sys.exit(1)
#
#    return args_list

def read_python_rc_file():
    """Read and parse the ~/.recdroidvid_rc.py file."""
    rc_path = os.path.abspath(os.path.join(os.path.expanduser("~"),
                              RECDROIDVID_PYTHON_RC_FILENAME))
    if not os.path.isfile(rc_path):
        return []

    module_filename = RECDROIDVID_PYTHON_RC_FILENAME[1:] # Remove the dot.
    module_name = RECDROIDVID_PYTHON_RC_FILENAME[1:-3] # Remove the dot and .py extension.
    with tempfile.TemporaryDirectory() as tmpdirname:
        shutil.copyfile(rc_path,
                os.path.join(tmpdirname, module_filename))
        sys.path.insert(1, tmpdirname)

        try:
            rc_options_module = __import__(module_name)
        except ImportError:
            print_error(f"\n\nERROR: RC file at '{rc_path}' raised an error on import:\n\n",
                  file=sys.stderr)
            raise
        finally:
            del sys.path[1]

    return rc_options_module.rdv_options




"""
A setuptools-based setup module.

See:
   https://packaging.python.org/en/latest/distributing.html
   https://github.com/pypa/sampleproject

Docs on the setup function kwargs:
   https://packaging.python.org/distributing/#setup-args

"""

import glob
import os.path
from setuptools import setup, find_packages
import codecs # Use a consistent encoding.

# Get the long description from the README.rst file.
current_dir = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(current_dir, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="recdroidvid",
    version="0.1.0", # major version, minor version, patch (see PEP440)
    description="Record and monitor video on android devices from computer (currently Linux via USB).",
    keywords=["android", "linux", "usb", "remote", "adb", "video", "movie", "record", "monitor"],
    install_requires=["wheel", "colorama"],
    python_requires=">=3.6",
    entry_points = {
        "console_scripts": ["recdroidvid = recdroidvid.recdroidvid_main:main"]
        },
    url="https://github.com/abarker/recdroidvid",

    license="MIT",
    classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        # Development Status: Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        #"Development Status :: 5 - Production/Stable",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        # uncomment if you test on these interpreters:
        # "Programming Language :: Python :: Implementation :: IronPython",
        # "Programming Language :: Python :: Implementation :: Jython",
        # "Programming Language :: Python :: Implementation :: Stackless",
        "Topic :: Utilities",
    ],

    # Settings usually the same.
    author="Allen Barker",
    author_email="Allen.L.Barker@gmail.com",
    include_package_data=True,
    zip_safe=False,

    # Automated stuff below.
    long_description=long_description,
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[os.path.splitext(os.path.basename(path))[0]
                                    for path in glob.glob("src/*.py")],
)




"""

Commands calling the Android Debug Bridge (ADB) to the moblie device.

"""

import sys
from time import sleep
from .utility_functions import run_local_cmd_blocking, print_error

args = None # Set globally from main() after command-line args are parsed.

def adb(cmd, *, print_cmd_str=True, exit_on_error=True):
    """Run the ADB command, printing out diagnostics.  Setting `return_output`
    returns the stdout of the command, but the command must be redirectable to
    a temp file.  Returned string is a direct read, with no splitting."""
    returncode, stdout, stderr = run_local_cmd_blocking(cmd, print_cmd_str=print_cmd_str,
                                                        print_cmd_prefix="ADB: ",
                                                        fail_on_nonzero_exit=False)
    if stderr.startswith("error: no devices"):
        print_error("\nERROR: No devices found, is the phone plugged in via USB,"
                    "\nwith access permission granted on the device?", file=sys.stderr)
        if exit_on_error:
            sys.exit(1)
        else:
            raise ADBException("No devices found.")
    elif returncode != 0:
        print_error(f"\nERROR: ADB command '{cmd}' returned nonzero exit status '{returncode}'."
                f"\nThe command's output follows:\n{stdout}\n{stderr}",
                file=sys.stderr)
        if exit_on_error:
            sys.exit(1)
        else:
            raise ADBException("Command returned nonzero exit status.")
    return stdout, stderr

def ls(path, also_hidden=False, extension_whitelist=None, print_cmd_str=True):
    """Run the ADB ls command and return the filenames time-sorted from oldest
    to newest.   If `all` is true the `-a` option to `ls` is used (which gets dotfiles
    too).  The `extension_whitelist` is an optional iterable of required file
    extensions such as `[".mp4"]`."""
    # NOTE NOTE: `adb shell ls` is DIFFERENT FROM `adb ls`, you need also hidden files with
    # `shell adb` to get `.pending....mp4` files, and there are still a few more in `shell ls`.
    if also_hidden:
        ls_list, ls_stderr = adb(f"adb shell ls -ctra {path}", print_cmd_str=print_cmd_str)
    else:
        ls_list, ls_stderr = adb(f"adb shell ls -ctr {path}", print_cmd_str=print_cmd_str)
    ls_list = ls_list.splitlines()

    if extension_whitelist:
        for e in extension_whitelist:
            ls_list = [f for f in ls_list if f.endswith(e)]
    return ls_list

def tap_screen(x, y):
    """Generate a screen tap at the given position."""
    #https://stackoverflow.com/questions/3437686/how-to-use-adb-to-send-touch-events-to-device-using-sendevent-command
    adb(f"adb shell input tap {x} {y}")

def force_stop_opencamera():
    """Issue a force-stop command to OpenCamera app.  Note this made the Google
    camera open by default afterward with camera button."""
    adb("adb shell am force-stop net.sourceforge.opencamera")

def tap_camera_button():
    """Tap the button in the camera to start it or stop it from recording."""
    adb(f"adb shell input keyevent 27")

def toggle_power():
    """Toggle the power.  See also the `device_wakeup` function."""
    adb("adb shell input keyevent KEYCODE_POWER")

def device_wakeup():
    """Issue an ADB wakeup command."""
    stdout, stderr = adb(f"adb shell input keyevent KEYCODE_WAKEUP")
    sleep(2)

def device_sleep(exit_on_error=True):
    """Issue an ADB sleep command."""
    stdout, stderr = adb(f"adb shell input keyevent KEYCODE_SLEEP", exit_on_error=exit_on_error)
    sleep(2)

def unlock_screen():
    """Swipes screen up, assuming no passcode."""
    # Note 82 is the menu key.
    #adb(f"adb shell input keyevent 82 && adb shell input keyevent 66")
    adb("adb shell input keyevent 3") # Simulate Home button to put down anything up.
    adb(f"adb shell input keyevent 82")
    sleep(1)

def open_video_camera():
    """Open the video camera, rear facing."""
    # Note that the -W option waits for the launch to complete.

    # NOTE: Below line fails sometimes when opening the menu instead of camera???
    #adb("adb shell am start -W net.sourceforge.opencamera --ei android.intent.extras.CAMERA_FACING 0")

    # Below depends on the default camera app, above forces OpenCamera.
    #adb(f"adb shell am start -W -a android.media.action.VIDEO_CAMERA --ei android.intent.extras.CAMERA_FACING 0")

    # This command seems to avoid opening in a menu, etc., for now....
    # https://android.stackexchange.com/questions/171490/start-application-from-adb
    # https://stackoverflow.com/questions/4567904/how-to-start-an-application-using-android-adb-tools
    adb("adb shell input keyevent 3") # Simulate Home button to put down anything up.
    adb(f"adb shell am start -W -n {args.camera_package_name[0]}/.MainActivity --ei android.intent.extras.CAMERA_FACING 0")
    sleep(1)

def directory_size_increasing(dirname, wait_secs=1):
    """Return true if the save directory is growing in size (i.e., file is being
    recorded there)."""
    DEBUG = False # Print commands to screen when debugging.
    first_du, stderr = adb(f"adb shell du {dirname}", print_cmd_str=DEBUG)
    first_du = first_du.split("\t")[0]
    sleep(wait_secs)
    second_du, stderr = adb(f"adb shell du {dirname}", print_cmd_str=DEBUG)
    second_du = second_du.split("\t")[0]
    return int(second_du) > int(first_du)

def pending_video_file_exists(dirname):
    """Return true if a filename starting with `.pending` is found in the directory.
    This is an implementation detail of OpenCamera, but can detect recording video
    in one call (unlike `directory_size_increasing`."""
    files = ls(dirname, also_hidden=True, print_cmd_str=False)
    return any(f.startswith(".pending") for f in files)

class ADBException(Exception):
    pass
